/*
 * Copyright (C) 2018 thomaskneisel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package pictureoptimizer;

import java.awt.image.BufferedImage;
import java.awt.Color;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.util.Random;
import java.util.zip.ZipOutputStream;
import java.io.InputStream;

/**
 * A single instance of an image
 *
 * @author thomaskneisel
 */
public class Individual {

	BufferedImage image;

	/**
	 * Create a randomised RGB image with given width and height
	 *
	 * @param width width of the image in pixels
	 * @param height height of the image in pixel
	 */
	public Individual(int width, int height) {
		Random rand = new Random();

		image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		for (int row = 0; row < width; row++) {
			for (int column = 0; column < height; column++) {
				int rgb = (new Color(rand.nextInt(255), rand.nextInt(255), rand.nextInt(255))).getRGB();
				image.setRGB(row, column, rgb);
			}
		}
	}

	/**
	 * Generate an image from two sources using cross over. To avoid cross over
	 * use the same source twice!
	 *
	 * @param p1 parent #1
	 * @param p2 parent #2
	 */
	public Individual(Individual p1, Individual p2) {
		Random rand = new Random();
		image = new BufferedImage(p1.image.getWidth(), p1.image.getHeight(),
				BufferedImage.TYPE_INT_RGB);

		// define the split point
		int rowBreak = rand.nextInt(p1.image.getHeight());
		int columnBreak = rand.nextInt(p1.image.getWidth());

		// copy source#1 until a column before the breakpoint 
		for (int column = 0; column < columnBreak; column++) {
			for (int row = 0; row < p1.image.getHeight(); row++) {
				image.setRGB(column, row, p1.image.getRGB(column, row));
			}
		}

		// copy the first part of the column with the breakpoint from source #1
		for (int row = 0; row < rowBreak; row++) {
			image.setRGB(columnBreak, row, p1.image.getRGB(columnBreak, row));
		}
		// copy the second part of the column with the breakpoint from source #2
		for (int row = rowBreak + 1; row < p1.image.getHeight(); row++) {
			image.setRGB(columnBreak, row, p2.image.getRGB(columnBreak, row));
		}

		// copy source#2 startting a column after the breakpoint 
		for (int column = columnBreak + 1; column < p1.image.getWidth(); column++) {
			for (int row = 0; row < p1.image.getHeight(); row++) {
				image.setRGB(column, row, p2.image.getRGB(column, row));
			}
		}
	}

	/**
	 * 
	 * @return reutrn the stored image 
	 */
	public BufferedImage getImage() {
		return image;
	}

	/**
	 * Mutate an image
	 *
	 * @param mutations number of mutations
	 */
	public void mutate(int mutations) {
		Random rand = new Random();
		for (int i = 0; i < mutations; i++) {
			int column = rand.nextInt(image.getWidth());
			int row = rand.nextInt(image.getHeight());
			int color = image.getRGB(column, row);
			color = color ^ (1 << rand.nextInt(24));
			image.setRGB(column, row, color);
		}
	}

	/**
	 * Compare the image with another image. Count the bits which have the same
	 * value in both images.
	 *
	 * @param target the image which is used for comparison
	 * @return number of bits with identical values
	 */
	int similarity(BufferedImage target) {
		int counter = 0;
		for (int row = 0; row < image.getHeight(); row++) {
			for (int column = 0; column < image.getWidth(); column++) {
				Color rgb1 = new Color(image.getRGB(column, row));
				Color rgb2 = new Color(target.getRGB(column, row));

				counter += BitCounter.countIdentical((int) rgb1.getRed(), (int) rgb2.getRed());
				counter += BitCounter.countIdentical((int) rgb1.getGreen(), (int) rgb2.getGreen());
				counter += BitCounter.countIdentical((int) rgb1.getBlue(), (int) rgb2.getBlue());
			}
		}
		return counter;
	}

	/**
	 * Calculate the fitness. Current implementation is counting the bits with
	 * identical values (see similarity())
	 *
	 * @param target the image which is used for comparison
	 * @return genetical fitness
	 */
	int fitness(BufferedImage target) {
		return similarity(target);
	}

	/**
	 * Write the image to a zip stream (BMP format used to prevent losses by
	 * compression). The file in the zip-stream needs to be opened before by the
	 * calling program
	 *
	 * @param zos zip-stream
	 * @throws IOException
	 */
	public void write(ZipOutputStream zos) throws IOException {
		try {
			ImageIO.write(image, "BMP", zos);
		} catch (IOException e) {
			throw (e);
		}
	}

	/**
	 * Read the image from a zip stream . The file in the zip-stream needs to be
	 * opened before by the calling program
	 *
	 * @param is zip-stream
	 * @throws IOException
	 */
	public Individual(InputStream is) throws IOException {
		image = ImageIO.read(is);
	}
}
