/*
 * Copyright (C) 2018 thomaskneisel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package pictureoptimizer;

import java.awt.image.BufferedImage;
import java.awt.Color;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.util.Random;
import java.lang.*;

/**
 *
 * @author thomaskneisel
 */
public class PictureOptimizer {

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {

        String projectname = "";
        String source = "";
        int imagesPerGeneration = 20;
        int mutations = 1;
        String selector = "";
        Population parents = null;
        Population childs = null;

        // parse commandline
		//*******************
        if ((args.length == 6) && args[0].equals("new")) {
			
            // 1) NEW - Create a new project
            //------------------------
            projectname = args[1];
            source = args[2];
            imagesPerGeneration = Integer.parseInt(args[3]);
            mutations = Integer.parseInt(args[4]);
            mutations = (mutations >= 0) ? mutations : 1;
            selector = args[5];

            // read the target image from file
            BufferedImage target = null;
            try {
                target = ImageIO.read(new File(source));
            } catch (IOException e) {
                System.err.println("ERROR: Could not read source image\n");
                System.exit(1);
            }
            int width = target.getWidth();
            int height = target.getHeight();

            System.out.println("Creating new project \"" + normaliseFilename(projectname, 0) + "\"...");
            System.out.println("\tSource File:\t\"" + source + "\"");
            System.out.println("\t\twidth:\t" + width + " px");
            System.out.println("\t\theight:\t" + height + " px");
            System.out.println("\tImages per generation:\t" + imagesPerGeneration);
            System.out.println("\tMutations per Generation:\t" + mutations);
            System.out.println("\tSelection Type:\t" + selector);

            parents = new Population(target, imagesPerGeneration, mutations, selector);
            try {
                parents.save(normaliseFilename(projectname, 0));
            } catch (IOException e) {
                System.err.println("ERROR: Could not write project file \"" + projectname + "\"");
                System.exit(1);
            }
        } else if ((args.length == 2) && args[0].equals("open")) {
            
			// 2) OPEN - Read an existing project from file
            //--------------------------------------
            projectname = args[1];
            try {
                parents = new Population(projectname);
            } catch (IOException e) {
                System.err.println("ERROR: Could not read project file \"" + projectname + "\"");
                System.exit(1);
            }

        } else {
			
            // 3) UNRECOGNISED COMMAND - line report an error
            //------------------------------------------------------
            System.err.println("ERROR: Invalid commandline\n");
            System.err.println("java PictureOptimizer new projectname targetimage images_per_generation mutation_rate selection_type");
            System.err.println("\tmutation_per_generation:  int > 0");
            System.err.println("\tselection_type: WinnerTakesItAll\n");
            System.err.println("java PictureOptimizer open projectname");
            System.exit(1);
        }

		int savecounter = 0;

		// the main loop
		//***************
		while (true) {
			// create a new generation
			System.out.println("Generation " + parents.getGeneration()
					+ "\t" + parents.getMaxSimilarity()
					+ "\t" + parents.getMaxFitness());
			childs = new Population(parents, parents.getImagesPerGeneration());

			// after 1000 generations, save to disk to miminimise data losses
			if (++savecounter == 1000) {
				savecounter = 0;
				try {
					System.out.println("\tsaving ...");
					childs.save(normaliseFilename(projectname, childs.getGeneration()));
				} catch (IOException e) {
					System.err.println("Could not save data");
					System.exit(1);
				}
			}
			parents = childs;
         }
    }

	/**
	 * create a filename in "canonical" form
	 * <ol>
	 *    <li>base filename (:= projectname) in lower case
	 *    </li>
	 *    <li> underscore "_"
	 *    </li>
	 *    <li>generation (9 digits)
	 *    </li>
	 *    <li>extension ".zip"
	 *    </li>
	 * </ol>
	 * 
	 * @param filename base of file name
	 * @param generation generation
	 * @return canonical form of filename
	 */
	public static String normaliseFilename(String filename, int generation) {
        String result = filename;
        if (result.substring(result.length() - 4, result.length()).toLowerCase().equals(".zip")) {
            result = result.substring(0, result.length() - 4);
        }
        while (Character.isDigit(result.charAt(result.length() - 1))) {
            result = result.substring(0, result.length() - 1);
        }
        while (result.charAt(result.length() - 1) == '_') {
            result = result.substring(0, result.length() - 1);
        }
        return result + "_" + String.format("%09d", generation) + ".zip";
    }
}
