/*
 * Copyright (C) 2018 thomaskneisel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package pictureoptimizer;

import java.awt.image.BufferedImage;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import java.util.zip.ZipFile;
import java.util.Enumeration;
import javax.imageio.ImageIO;
import java.util.Random;

/**
 * A population as a collection of Individuals
 *
 * @author thomaskneisel
 */
public class Population {

	int generation;
	int mutations;
	BufferedImage target;
	Individual individuals[];
	int similarity[];
	int fitness[];
	double rouletteWheel[];
	Selection selector;

	/**
	 * Create a Population of individuals as pure randomised images
	 *
	 * @param target_ target of the optimisation
	 * @param n size of the population
	 */
	Population(BufferedImage target_, int n, int mutations_, String selector_) {
		generation = 0;
		mutations = mutations_;
		selector = Selection.createSelection(selector_);

		target = target_;
		individuals = new Individual[n];
		for (int i = 0; i < n; i++) {
			individuals[i] = new Individual(target.getWidth(), target.getHeight());
		}
		createStatistic();
	}

	/**
	 * Create a population from another population using evolutionary algorithms
	 *
	 * @param parents the ancestors
	 * @param n size of the population
	 */
	Population(Population parents, int n) {
		generation = parents.generation + 1;
		mutations = parents.mutations;
		selector = parents.selector;

		target = parents.target;
		individuals = new Individual[n];

		// select two parents, do crossover and apply mutations
		for (int i = 0; i < n; i++) {
			Individual p1 = parents.selectParent();
			Individual p2 = parents.selectParent();
			individuals[i] = new Individual(p1, p2);
			individuals[i].mutate(mutations);
		}
		createStatistic();
	}

	/**
	 * Generate statistical data about the population
	 */
	private void createStatistic() {
		// Calculate similarity with target
		similarity = new int[individuals.length];
		for (int i = 0; i < individuals.length; i++) {
			similarity[i] = individuals[i].similarity(target);
		}

		// Calculate similarity with fitness of all members
		fitness = new int[individuals.length];
		for (int i = 0; i < individuals.length; i++) {
			fitness[i] = individuals[i].fitness(target);
		}

		// sort individuals by fitness in decreasing order
		// (I know I am idiot, becuase I am using simple bubble sort)
		//
		// fitness[0]: maximum
		// fitness[individuals.length / 2] : median
		// fitness[individuals.length - 1] : minimum
		boolean unsorted;
		do {
			unsorted = false;
			for (int i = 0; i < individuals.length - 1; i++) {
				if (fitness[i] < fitness[i + 1]) {
					Individual temp1 = individuals[i + 1];
					individuals[i + 1] = individuals[i];
					individuals[i] = temp1;

					int temp2 = fitness[i + 1];
					fitness[i + 1] = fitness[i];
					fitness[i] = temp2;

					temp2 = similarity[i + 1];
					similarity[i + 1] = similarity[i];
					similarity[i] = temp2;

					unsorted = true;
					break;
				}
			}
		} while (unsorted);

		for (int i = 0; i < individuals.length; i++) {
			similarity[i] = individuals[i].similarity(target);
		}

		// create the weighted roulettewheel for selection
		rouletteWheel = selector.normalise(fitness);
	}

	/**
	 * 
	 * @return number of individuals in the population 
	 */
	public int getImagesPerGeneration() {
		return individuals.length;
	}

	/**
	 * 
	 * @return number of current generation 
	 */
	public int getGeneration() {
		return generation;
	}
	/**
	 * 
	 * @return best fitness in population 
	 */
	public int getMaxFitness() {
		return fitness[0];
	}
	
	/**
	 * 
	 * @return highest similarity with target in population 
	 */
	public int getMaxSimilarity() {
		return similarity[0];
	}

	public int getMedianFitness() {
		return fitness[individuals.length / 2];
	}

	/**
	 * 
	 * @return average of fitness 
	 */
	public int getAverageFitness() {
		long totalFitness = 0;
		for (int i = 0; i < individuals.length; i++) {
			totalFitness += (long) fitness[i];
		}
		return (int) (totalFitness / (long) individuals.length);
	}

	/**
	 * 
	 * @return lowest fitness value 
	 */
	public int getMinFitness() {
		return fitness[individuals.length - 0];
	}

	/**
	 * 
	 * @return best image in population
	 */
	public BufferedImage getBestImage() {
		return individuals[0].getImage();
	}

	/**
	 * Select an member of the population by weighted roulette wheel
	 *
	 * @return a member of the population according selection process
	 */
	public Individual selectParent() {
		int i;
		Random rand = new Random();
		double r = Math.random();
		for (i = 0; i < individuals.length; i++) {
			if (r < rouletteWheel[i]) {
				break;
			}
		}
		return individuals[i];
	}

	/**
	 * Save the entire population in a zip file
	 *
	 * @param filename name of the zip archive (in "canonical" form)
	 * @throws IOException in case of error
	 */
	public void save(String filename) throws IOException {
		// Everything will be put in a zip file
		ZipOutputStream zos = null;
		try {
			// create zipfile
			zos = new ZipOutputStream(new FileOutputStream(filename));

			// save project parameters in a simple CSV-file
			zos.putNextEntry(new ZipEntry("basedata.txt"));
			PrintWriter printer = new PrintWriter(zos);
			printer.println("version;" + "21-Jan-2016");
			printer.println("generation;" + generation);
			printer.println("mutationRate;" + mutations);
			printer.println("size;" + individuals.length);
			printer.println("selection;" + selector.info());
			printer.flush();
			zos.closeEntry();

			// save statistical data in a simple CSV-file
			zos.putNextEntry(new ZipEntry("statistic.txt"));
			printer = new PrintWriter(zos);
			printer.println("n;similarity;fitness;rouletteWheel");
			for (int i = 0; i < individuals.length; i++) {
				printer.println(i + ";"
						+ similarity[i] + ";"
						+ fitness[i] + ";"
						+ rouletteWheel[i]);
			}
			printer.flush();
			zos.closeEntry();

			// save the target picture
			zos.putNextEntry(new ZipEntry("target.bmp"));
			ImageIO.write(target, "BMP", zos);
			zos.closeEntry();

			// save the entire population pictures
			for (int i = 0; i < individuals.length; i++) {
				zos.putNextEntry(new ZipEntry("member" + i + ".bmp"));
				individuals[i].write(zos);
				zos.closeEntry();
			}
			
		// some error handling	
		} catch (IOException e) {
			throw (e);
		} finally {
			if (zos != null) {
				try {
					zos.closeEntry();
					zos.close();
				} catch (IOException e) {
					throw (e);
				}
			}
		}
	}

	/**
	 * Read a population from a zip file.
	 * (data need to stored as a file, reading from stream is not supported)
	 *
	 * @param filename name of zip archive (in "canonical" form)
	 * @throws IOException in case of error
	 */
	public Population(String filename) throws IOException {
		ZipFile file = new ZipFile(filename);
		try {
			// read the projects basedata
			readBaseData(findZipEntryByName(file, "basedata.txt"));
			// ignore statistical data in zip file
			// these data are calculated after all data is loaded
			;

			// read the target picture
			target = ImageIO.read(findZipEntryByName(file, "target.bmp"));

			// read all members of the current generation
			for (int i = 0; i < individuals.length; i++) {
				individuals[i] = new Individual(findZipEntryByName(file, "member" + i + ".bmp"));

			}

			// refresh statistic information
			createStatistic();
		} catch (IOException e) {
			throw e;
		} finally {
			file.close();
		}
	}

	/**
	 * Jump to the beginning of a given file in a zip-archive file. The file is
	 * identified by its name.
	 *
	 * @param file zip-archive
	 * @param entryName the name of the store in the zip-archive
	 * @return input stream for this zip file entry, null if element could not
	 * be found or when an IO error occurs
	 */
	private InputStream findZipEntryByName(ZipFile file, String entryName) throws IOException {
		try {
			// retrieve information about all files stored in the archive
			Enumeration<? extends ZipEntry> entries = file.entries();

			// iterate through the archive and try to find the requested entry
			// return an input stream associated with the entry or null
			// if entry was not found
			boolean dataFound = false;
			while (entries.hasMoreElements()) {
				ZipEntry entry = entries.nextElement();
				if (entry.getName().equals(entryName)) {
					return file.getInputStream(entry);
				}
			}
			throw new IOException("Could not find zip-file entry " + entryName);
		} catch (IOException e) {
			throw e;
		}
	}

	/**
	 * Read generations basedata from csv stream
	 *
	 * @param is input stream of the basedata.txt file
	 * @throws IOException in case of IO error or when keywords were not found
	 */
	private void readBaseData(InputStream is) throws IOException {
		BufferedReader r = new BufferedReader(new InputStreamReader(is));
		String s;
		HashMap<String, String> map = new HashMap();

		// first we read the file, split each line in 
		// keyword and respective value (as String).
		// keyword and value are stored in a hashmap
		do {
			try {
				s = r.readLine();
				if (s == null) {
					break;
				}
				String temp[] = s.split(";");
				if (temp.length == 2) {
					map.put(temp[0].toLowerCase(), temp[1]);
				}
			} catch (IOException e) {
				throw e;
			}
		} while (true);

		String temp;
		int n;

		// retrieve the values from hashmap and store them in class variable
		temp = map.get("generation");
		if (temp != null) {
			generation = Integer.parseInt(temp);
		} else {
			throw new IOException("basedata.txt does not contain key \"generation\"");
		}

		temp = map.get("size");
		if (temp != null) {
			n = Integer.parseInt(temp);
		} else {
			throw new IOException("basedata.txt does not contain key \"size\"");
		}

		temp = map.get("mutationrate");
		if (temp != null) {
			mutations = Integer.parseInt(temp);
		} else {
			throw new IOException("basedata.txt does not contain key \"mutationrate\"");
		}

		temp = map.get("selection");
		if (temp != null) {
			selector = Selection.createSelection(temp);
		} else {
			throw new IOException("basedata.txt does not contain key \"selection\"");
		}

		// create an array according the number of individuals in the generation
		individuals = new Individual[n];
	}
}
